# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from dagorama.api import api_pb2 as dagorama_dot_api_dot_api__pb2


class DagoramaStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.CreateWorker = channel.unary_unary(
                '/main.Dagorama/CreateWorker',
                request_serializer=dagorama_dot_api_dot_api__pb2.WorkerConfigurationMessage.SerializeToString,
                response_deserializer=dagorama_dot_api_dot_api__pb2.WorkerMessage.FromString,
                )
        self.CreateInstance = channel.unary_unary(
                '/main.Dagorama/CreateInstance',
                request_serializer=dagorama_dot_api_dot_api__pb2.InstanceConfigurationMessage.SerializeToString,
                response_deserializer=dagorama_dot_api_dot_api__pb2.InstanceMessage.FromString,
                )
        self.CreateNode = channel.unary_unary(
                '/main.Dagorama/CreateNode',
                request_serializer=dagorama_dot_api_dot_api__pb2.NodeConfigurationMessage.SerializeToString,
                response_deserializer=dagorama_dot_api_dot_api__pb2.NodeMessage.FromString,
                )
        self.Ping = channel.unary_unary(
                '/main.Dagorama/Ping',
                request_serializer=dagorama_dot_api_dot_api__pb2.WorkerMessage.SerializeToString,
                response_deserializer=dagorama_dot_api_dot_api__pb2.PongMessage.FromString,
                )
        self.GetNode = channel.unary_unary(
                '/main.Dagorama/GetNode',
                request_serializer=dagorama_dot_api_dot_api__pb2.NodeRetrieveMessage.SerializeToString,
                response_deserializer=dagorama_dot_api_dot_api__pb2.NodeMessage.FromString,
                )
        self.GetWork = channel.unary_unary(
                '/main.Dagorama/GetWork',
                request_serializer=dagorama_dot_api_dot_api__pb2.WorkerMessage.SerializeToString,
                response_deserializer=dagorama_dot_api_dot_api__pb2.NodeMessage.FromString,
                )
        self.SubmitWork = channel.unary_unary(
                '/main.Dagorama/SubmitWork',
                request_serializer=dagorama_dot_api_dot_api__pb2.WorkCompleteMessage.SerializeToString,
                response_deserializer=dagorama_dot_api_dot_api__pb2.NodeMessage.FromString,
                )
        self.SubmitFailure = channel.unary_unary(
                '/main.Dagorama/SubmitFailure',
                request_serializer=dagorama_dot_api_dot_api__pb2.WorkFailedMessage.SerializeToString,
                response_deserializer=dagorama_dot_api_dot_api__pb2.NodeMessage.FromString,
                )


class DagoramaServicer(object):
    """Missing associated documentation comment in .proto file."""

    def CreateWorker(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateInstance(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateNode(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Ping(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetNode(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetWork(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SubmitWork(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SubmitFailure(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_DagoramaServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'CreateWorker': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateWorker,
                    request_deserializer=dagorama_dot_api_dot_api__pb2.WorkerConfigurationMessage.FromString,
                    response_serializer=dagorama_dot_api_dot_api__pb2.WorkerMessage.SerializeToString,
            ),
            'CreateInstance': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateInstance,
                    request_deserializer=dagorama_dot_api_dot_api__pb2.InstanceConfigurationMessage.FromString,
                    response_serializer=dagorama_dot_api_dot_api__pb2.InstanceMessage.SerializeToString,
            ),
            'CreateNode': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateNode,
                    request_deserializer=dagorama_dot_api_dot_api__pb2.NodeConfigurationMessage.FromString,
                    response_serializer=dagorama_dot_api_dot_api__pb2.NodeMessage.SerializeToString,
            ),
            'Ping': grpc.unary_unary_rpc_method_handler(
                    servicer.Ping,
                    request_deserializer=dagorama_dot_api_dot_api__pb2.WorkerMessage.FromString,
                    response_serializer=dagorama_dot_api_dot_api__pb2.PongMessage.SerializeToString,
            ),
            'GetNode': grpc.unary_unary_rpc_method_handler(
                    servicer.GetNode,
                    request_deserializer=dagorama_dot_api_dot_api__pb2.NodeRetrieveMessage.FromString,
                    response_serializer=dagorama_dot_api_dot_api__pb2.NodeMessage.SerializeToString,
            ),
            'GetWork': grpc.unary_unary_rpc_method_handler(
                    servicer.GetWork,
                    request_deserializer=dagorama_dot_api_dot_api__pb2.WorkerMessage.FromString,
                    response_serializer=dagorama_dot_api_dot_api__pb2.NodeMessage.SerializeToString,
            ),
            'SubmitWork': grpc.unary_unary_rpc_method_handler(
                    servicer.SubmitWork,
                    request_deserializer=dagorama_dot_api_dot_api__pb2.WorkCompleteMessage.FromString,
                    response_serializer=dagorama_dot_api_dot_api__pb2.NodeMessage.SerializeToString,
            ),
            'SubmitFailure': grpc.unary_unary_rpc_method_handler(
                    servicer.SubmitFailure,
                    request_deserializer=dagorama_dot_api_dot_api__pb2.WorkFailedMessage.FromString,
                    response_serializer=dagorama_dot_api_dot_api__pb2.NodeMessage.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'main.Dagorama', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Dagorama(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def CreateWorker(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/main.Dagorama/CreateWorker',
            dagorama_dot_api_dot_api__pb2.WorkerConfigurationMessage.SerializeToString,
            dagorama_dot_api_dot_api__pb2.WorkerMessage.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateInstance(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/main.Dagorama/CreateInstance',
            dagorama_dot_api_dot_api__pb2.InstanceConfigurationMessage.SerializeToString,
            dagorama_dot_api_dot_api__pb2.InstanceMessage.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateNode(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/main.Dagorama/CreateNode',
            dagorama_dot_api_dot_api__pb2.NodeConfigurationMessage.SerializeToString,
            dagorama_dot_api_dot_api__pb2.NodeMessage.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Ping(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/main.Dagorama/Ping',
            dagorama_dot_api_dot_api__pb2.WorkerMessage.SerializeToString,
            dagorama_dot_api_dot_api__pb2.PongMessage.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetNode(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/main.Dagorama/GetNode',
            dagorama_dot_api_dot_api__pb2.NodeRetrieveMessage.SerializeToString,
            dagorama_dot_api_dot_api__pb2.NodeMessage.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetWork(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/main.Dagorama/GetWork',
            dagorama_dot_api_dot_api__pb2.WorkerMessage.SerializeToString,
            dagorama_dot_api_dot_api__pb2.NodeMessage.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SubmitWork(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/main.Dagorama/SubmitWork',
            dagorama_dot_api_dot_api__pb2.WorkCompleteMessage.SerializeToString,
            dagorama_dot_api_dot_api__pb2.NodeMessage.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SubmitFailure(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/main.Dagorama/SubmitFailure',
            dagorama_dot_api_dot_api__pb2.WorkFailedMessage.SerializeToString,
            dagorama_dot_api_dot_api__pb2.NodeMessage.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from base64 import b64encode
from hashlib import md5
from inspect import getfullargspec
from marshal import dumps
from typing import Any, Callable

import pkg_resources


def get_explicit_dependencies():
    """
    Gets one layer of dependencies, does not attempt to recurse down the
    dependency tree.

    """
    return [
        (package.project_name, package.version)
        for package in pkg_resources.working_set
    ]

def calculate_function_hash(fn: Callable[..., Any], include_package_versions: bool = False) -> str:
    """
    :param include_package_versions: Will also include the package versions
        as part of the hash. This helps ensures there is strict environment
        equality between the deployment and runner but is likely overkill for
        most situations.

    """
    argspec = getfullargspec(fn)
    runtime_code_raw = dumps(
        dict(
            name=fn.__name__,
            code=fn.__code__.co_code,
            constants=fn.__code__.co_consts,
            args=argspec.args,
            kwargs=argspec.kwonlyargs,
        )
    )

    if include_package_versions:
        dependencies = get_explicit_dependencies()
        dependencies = sorted(dependencies, key=lambda x: x[0])

        runtime_code_raw += dumps(dependencies)

    return b64encode(md5(runtime_code_raw).digest()).decode()



from inspect import getsourcelines
from multiprocessing.sharedctypes import Value
from tempfile import NamedTemporaryFile
from typing import no_type_check

from mypy import api as mypy_api


# This will be typechecked in a separate process
# We want it to create an error
@no_type_check
def logic_with_unnecessary_argument():
    from dagorama.definition import DAGDefinition, dagorama

    class CustomDag(DAGDefinition):
        @dagorama
        def entrypoint(self, number: int):
            self.linear_continue(number, 10)
            return number + 1

        def linear_continue(self, number: int):
            return number + 1


def test_unnecessary_argument_typecheck():
    with NamedTemporaryFile() as file:
        # Typechecking is static so we need to isolate our erroring code to a new file
        file.write("\n".join(getsourcelines(logic_with_unnecessary_argument)[0]).encode())
        file.seek(0)

        type_check_report, _, error_code = mypy_api.run([file.name])

    assert 'Too many arguments for "linear_continue" of "CustomDag"' in type_check_report

    assert error_code == 1



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class WorkerConfigurationMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    EXCLUDEQUEUES_FIELD_NUMBER: builtins.int
    INCLUDEQUEUES_FIELD_NUMBER: builtins.int
    QUEUETOLERATIONS_FIELD_NUMBER: builtins.int
    @property
    def excludeQueues(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    @property
    def includeQueues(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    @property
    def queueTolerations(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        excludeQueues: collections.abc.Iterable[builtins.str] | None = ...,
        includeQueues: collections.abc.Iterable[builtins.str] | None = ...,
        queueTolerations: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["excludeQueues", b"excludeQueues", "includeQueues", b"includeQueues", "queueTolerations", b"queueTolerations"]) -> None: ...

global___WorkerConfigurationMessage = WorkerConfigurationMessage

@typing_extensions.final
class WorkerMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    IDENTIFIER_FIELD_NUMBER: builtins.int
    identifier: builtins.str
    def __init__(
        self,
        *,
        identifier: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["identifier", b"identifier"]) -> None: ...

global___WorkerMessage = WorkerMessage

@typing_extensions.final
class InstanceConfigurationMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    IDENTIFIER_FIELD_NUMBER: builtins.int
    identifier: builtins.str
    def __init__(
        self,
        *,
        identifier: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["identifier", b"identifier"]) -> None: ...

global___InstanceConfigurationMessage = InstanceConfigurationMessage

@typing_extensions.final
class InstanceMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    IDENTIFIER_FIELD_NUMBER: builtins.int
    identifier: builtins.str
    def __init__(
        self,
        *,
        identifier: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["identifier", b"identifier"]) -> None: ...

global___InstanceMessage = InstanceMessage

@typing_extensions.final
class NodeConfigurationMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    IDENTIFIER_FIELD_NUMBER: builtins.int
    FUNCTIONNAME_FIELD_NUMBER: builtins.int
    FUNCTIONHASH_FIELD_NUMBER: builtins.int
    QUEUENAME_FIELD_NUMBER: builtins.int
    TAINTNAME_FIELD_NUMBER: builtins.int
    ARGUMENTS_FIELD_NUMBER: builtins.int
    SOURCEIDS_FIELD_NUMBER: builtins.int
    INSTANCEID_FIELD_NUMBER: builtins.int
    RETRY_FIELD_NUMBER: builtins.int
    identifier: builtins.str
    functionName: builtins.str
    functionHash: builtins.str
    queueName: builtins.str
    taintName: builtins.str
    arguments: builtins.bytes
    @property
    def sourceIds(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    instanceId: builtins.str
    @property
    def retry(self) -> global___RetryMessage:
        """Retry customization"""
    def __init__(
        self,
        *,
        identifier: builtins.str = ...,
        functionName: builtins.str = ...,
        functionHash: builtins.str = ...,
        queueName: builtins.str = ...,
        taintName: builtins.str = ...,
        arguments: builtins.bytes = ...,
        sourceIds: collections.abc.Iterable[builtins.str] | None = ...,
        instanceId: builtins.str = ...,
        retry: global___RetryMessage | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["retry", b"retry"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["arguments", b"arguments", "functionHash", b"functionHash", "functionName", b"functionName", "identifier", b"identifier", "instanceId", b"instanceId", "queueName", b"queueName", "retry", b"retry", "sourceIds", b"sourceIds", "taintName", b"taintName"]) -> None: ...

global___NodeConfigurationMessage = NodeConfigurationMessage

@typing_extensions.final
class NodeMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    IDENTIFIER_FIELD_NUMBER: builtins.int
    FUNCTIONNAME_FIELD_NUMBER: builtins.int
    FUNCTIONHASH_FIELD_NUMBER: builtins.int
    QUEUENAME_FIELD_NUMBER: builtins.int
    TAINTNAME_FIELD_NUMBER: builtins.int
    ARGUMENTS_FIELD_NUMBER: builtins.int
    RESOLVEDVALUE_FIELD_NUMBER: builtins.int
    SOURCES_FIELD_NUMBER: builtins.int
    COMPLETED_FIELD_NUMBER: builtins.int
    INSTANCEID_FIELD_NUMBER: builtins.int
    identifier: builtins.str
    functionName: builtins.str
    functionHash: builtins.str
    queueName: builtins.str
    """Queue parameterization"""
    taintName: builtins.str
    arguments: builtins.bytes
    resolvedValue: builtins.bytes
    @property
    def sources(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___NodeMessage]:
        """Nodes that this node depends on
        By the time that this message is sent back to client callers
        these values should be fully populated
        """
    completed: builtins.bool
    instanceId: builtins.str
    def __init__(
        self,
        *,
        identifier: builtins.str = ...,
        functionName: builtins.str = ...,
        functionHash: builtins.str = ...,
        queueName: builtins.str = ...,
        taintName: builtins.str = ...,
        arguments: builtins.bytes = ...,
        resolvedValue: builtins.bytes = ...,
        sources: collections.abc.Iterable[global___NodeMessage] | None = ...,
        completed: builtins.bool = ...,
        instanceId: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["arguments", b"arguments", "completed", b"completed", "functionHash", b"functionHash", "functionName", b"functionName", "identifier", b"identifier", "instanceId", b"instanceId", "queueName", b"queueName", "resolvedValue", b"resolvedValue", "sources", b"sources", "taintName", b"taintName"]) -> None: ...

global___NodeMessage = NodeMessage

@typing_extensions.final
class RetryMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MAXATTEMPTS_FIELD_NUMBER: builtins.int
    STATICINTERVAL_FIELD_NUMBER: builtins.int
    EXPONENTIALBASE_FIELD_NUMBER: builtins.int
    maxAttempts: builtins.int
    staticInterval: builtins.int
    exponentialBase: builtins.int
    def __init__(
        self,
        *,
        maxAttempts: builtins.int = ...,
        staticInterval: builtins.int = ...,
        exponentialBase: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["exponentialBase", b"exponentialBase", "maxAttempts", b"maxAttempts", "staticInterval", b"staticInterval"]) -> None: ...

global___RetryMessage = RetryMessage

@typing_extensions.final
class PongMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LASTPING_FIELD_NUMBER: builtins.int
    lastPing: builtins.int
    def __init__(
        self,
        *,
        lastPing: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["lastPing", b"lastPing"]) -> None: ...

global___PongMessage = PongMessage

@typing_extensions.final
class WorkCompleteMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    INSTANCEID_FIELD_NUMBER: builtins.int
    NODEID_FIELD_NUMBER: builtins.int
    WORKERID_FIELD_NUMBER: builtins.int
    RESULT_FIELD_NUMBER: builtins.int
    instanceId: builtins.str
    nodeId: builtins.str
    workerId: builtins.str
    result: builtins.bytes
    def __init__(
        self,
        *,
        instanceId: builtins.str = ...,
        nodeId: builtins.str = ...,
        workerId: builtins.str = ...,
        result: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["instanceId", b"instanceId", "nodeId", b"nodeId", "result", b"result", "workerId", b"workerId"]) -> None: ...

global___WorkCompleteMessage = WorkCompleteMessage

@typing_extensions.final
class WorkFailedMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    INSTANCEID_FIELD_NUMBER: builtins.int
    NODEID_FIELD_NUMBER: builtins.int
    WORKERID_FIELD_NUMBER: builtins.int
    TRACEBACK_FIELD_NUMBER: builtins.int
    instanceId: builtins.str
    nodeId: builtins.str
    workerId: builtins.str
    traceback: builtins.str
    def __init__(
        self,
        *,
        instanceId: builtins.str = ...,
        nodeId: builtins.str = ...,
        workerId: builtins.str = ...,
        traceback: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["instanceId", b"instanceId", "nodeId", b"nodeId", "traceback", b"traceback", "workerId", b"workerId"]) -> None: ...

global___WorkFailedMessage = WorkFailedMessage

@typing_extensions.final
class NodeRetrieveMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    INSTANCEID_FIELD_NUMBER: builtins.int
    IDENTIFIER_FIELD_NUMBER: builtins.int
    instanceId: builtins.str
    identifier: builtins.str
    def __init__(
        self,
        *,
        instanceId: builtins.str = ...,
        identifier: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["identifier", b"identifier", "instanceId", b"instanceId"]) -> None: ...

global___NodeRetrieveMessage = NodeRetrieveMessage



from dagorama.decorators import dagorama
from dagorama.definition import (DAGDefinition, DAGInstance,
                                 generate_instance_id)
from dagorama.runner import execute


class SampleDefinition(DAGDefinition):
    @dagorama().syncfn
    def entrypoint(self, *args, **kwargs):
        self.standard_function()
        assert isinstance(self, DAGInstance)

    def standard_function(self):
        assert isinstance(self, DAGInstance)
        return 1

def test_instance_injection(broker):
    """
    Ensure the instance is injected into the runtime

    """
    definition = SampleDefinition()
    instance = DAGInstance(generate_instance_id(), definition)
    instance.entrypoint()

    execute(infinite_loop=False, catch_exceptions=False)



from asyncio import run
from contextlib import contextmanager
from inspect import isawaitable
from multiprocessing import Process
from pickle import dumps, loads
from threading import Thread
from time import sleep
from traceback import format_exc

import grpc

import dagorama.api.api_pb2 as pb2
import dagorama.api.api_pb2_grpc as pb2_grpc
from dagorama.code_signature import calculate_function_hash
from dagorama.definition import dagorama_context
from dagorama.inspection import resolve_promises
from dagorama.models.arguments import DAGArguments
from dagorama.serializer import name_to_function


class CodeMismatchException(Exception):
    """
    If raised, the queued DAGNode and the current runner are using
    different versions of the function code.

    """
    pass


def schedule_ping(
    context: pb2_grpc.DagoramaStub,
    worker: pb2.WorkerMessage,
    interval: int = 30,
):
    """
    Send a ping to the server to keep the connection alive.
    """
    while True:
        context.Ping(worker)
        sleep(interval)


async def execute_async(
    exclude_queues: list | None = None,
    include_queues: list | None = None,
    queue_tolerations: list | None = None,
    infinite_loop: bool = True,
    catch_exceptions: bool = True,
):
    with dagorama_context() as context:
        worker = context.CreateWorker(
            pb2.WorkerConfigurationMessage(
                excludeQueues=exclude_queues or [],
                includeQueues=include_queues or [],
                queueTolerations=queue_tolerations or [],
            )
        )

        Thread(
            target=schedule_ping,
            args=(context, worker),
            # Stop the thread when the rest of the process finishes
            daemon=True,
        ).start()

        while True:
            try:
                next_item = context.GetWork(worker)
            except grpc.RpcError as e:
                if e.details() == "no work available":
                    if not infinite_loop:
                        return
                    print("No work available, polling in 1s...")
                    sleep(1)
                    continue
                else:
                    raise

            arguments = DAGArguments.from_server_bytes(next_item.arguments)

            resolved_values = {
                source.identifier: loads(source.resolvedValue)
                for source in next_item.sources
            }

            # TODO: Add actual resolution
            print("------")
            print(next_item)
            print(arguments)
            print("resolved values", resolved_values)
            resolved_args = resolve_promises(arguments.calltime_args, resolved_values)
            resolved_kwargs = resolve_promises(arguments.calltime_kwargs, resolved_values)
            print("Resolved", resolved_args)
            print("Resolved", resolved_kwargs)

            # Since this function was queued as part of the worker, we can assume that it will
            # be a wrapped @dagorama function - it will therefore be a standard callable without
            # async since it takes care of the client-side async logic internally.
            resolved_fn = name_to_function(next_item.functionName, next_item.instanceId)
            print("RESOLVED FN", resolved_fn)

            # Ensure that we have the correct local version of the function
            print("FOUND FN", resolved_fn)
            #print("COMPARE VALUES", calculate_function_hash(resolved_fn.original_fn), next_item.functionHash)
            if calculate_function_hash(resolved_fn.original_fn) != next_item.functionHash:
                raise CodeMismatchException()

            if catch_exceptions:
                try:
                    result = resolved_fn(*resolved_args, greedy_execution=True, **resolved_kwargs)
                    if isawaitable(result):
                        result = await result
                except Exception as e:
                    traceback = format_exc()
                    print("Exception encountered, reporting to broker:", e, traceback)
                    context.SubmitFailure(
                        pb2.WorkFailedMessage(
                            instanceId=next_item.instanceId,
                            nodeId=next_item.identifier,
                            workerId=worker.identifier,
                            traceback=traceback
                        )
                    )
                    continue
            else:
                result = resolved_fn(*resolved_args, greedy_execution=True, **resolved_kwargs)
                if isawaitable(result):
                    result = await result

            context.SubmitWork(
                pb2.WorkCompleteMessage(
                    instanceId=next_item.instanceId,
                    nodeId=next_item.identifier,
                    workerId=worker.identifier,
                    result=dumps(result),
                )
            )


def execute(
    exclude_queues: list | None = None,
    include_queues: list | None = None,
    queue_tolerations: list | None = None,
    infinite_loop: bool = True,
    catch_exceptions: bool = True,
):
    """
    Run the execute function in a synchronous manner. Assumes no runloop
    is already running or asyncio will raise an error.

    """
    return run(
        execute_async(
            exclude_queues=exclude_queues,
            include_queues=include_queues,
            queue_tolerations=queue_tolerations,
            infinite_loop=infinite_loop,
            catch_exceptions=catch_exceptions,
        )
    )


@contextmanager
def launch_workers(n: int = 1):
    """
    Helper function to spawn multiple workers without having to launch
    them in separate shell sessions.

    Mostly intended for unit testing.

    """
    workers = [Process(target=execute) for _ in range(n)]
    for worker in workers:
        worker.start()

    yield

    for worker in workers:
        worker.terminate()




from importlib import import_module
from inspect import isclass
from uuid import UUID

from dagorama.definition import DAGDefinition, DAGInstance


def function_to_name(func):
    if isclass(func):
        module_name = getattr(func, "__module__")
        class_path = getattr(func, "__name__")
    elif isinstance(func, (DAGDefinition, DAGInstance)):
        return function_to_name(func.__class__)
    else:
        module_name = func.__globals__["__name__"]
        class_path = func.__qualname__

    return f"{module_name}:{class_path}"


DEFINITION_CACHE = {}


def register_definition(definition: DAGDefinition):
    global DEFINITION_CACHE
    DEFINITION_CACHE[function_to_name(definition)] = definition


def name_to_function(name: str, instance_id: UUID):
    """
    Takes a function name and attempts to convert it into a local callable.
    If the function is part of a class that hasn't already been registered as
    via `register_definition`, will attempt to instantiate it with no initial
    parameters.

    If your class needs to be instantiated with runtime variables, make sure you
    register it before the worker begins executing.

    """
    # Can also try to sniff for function names throughout entire global namespace, but this
    # is still conditioned on name conventions
    package_name, fn_path = name.split(":")
    fn_path_components = fn_path.split(".")

    try:
        mod = import_module(package_name)
    except ModuleNotFoundError:
        raise ValueError(f"Could not import function with path: {name}")

    for i, component in enumerate(fn_path_components):
        # Get class from cached globals, if possible
        component_name = f"{package_name}:{'.'.join(fn_path_components[:i+1])}"
        if component_name in DEFINITION_CACHE:
            mod = DEFINITION_CACHE[component_name]  # type: ignore
        else:
            if not hasattr(mod, component):
                raise ValueError(f"Unable to resolve dagorama function: {name}")

            mod = getattr(mod, component)

            if isclass(mod):
                # Assume we can instantiate the class with no parameters
                mod = mod()

            # We should only register DAGDefinitions, since these are the required
            # abstraction layer for 
            if isinstance(mod, DAGDefinition):
                register_definition(mod)

        # We should re-wrap this instance with the context that we know
        # from the message
        if isinstance(mod, DAGDefinition):
            mod = DAGInstance(instance_id, mod) # type: ignore

    return mod



from inspect import getfullargspec

from dagorama.decorators import dagorama
from dagorama.definition import DAGDefinition, resolve
from dagorama.runner import execute, execute_async
from dagorama.serializer import register_definition


class CustomDAG(DAGDefinition):
    def __init__(self, number: int):
        self.number = number

    @dagorama().syncfn
    def entrypoint(self):
        return self.number


def test_runner_custom_init(broker):
    """
    Ensure that we are able to register a custom version
    of the local DAGDefinition that we want to leverage

    """

    dag = CustomDAG(number=10)
    dag_instance, dag_result = dag()

    register_definition(dag)
    execute(infinite_loop=False, catch_exceptions=False)

    assert resolve(dag_instance, dag_result) == 10


def test_sync_async_parity():
    """
    Ensure the sync and async runner functions have the same API.

    """
    sync_spec = getfullargspec(execute)
    async_spec = getfullargspec(execute_async)
    assert sync_spec.args == async_spec.args



import pytest

from dagorama.decorators import dagorama
from dagorama.definition import DAGDefinition, resolve
from dagorama.retry import ExponentialRetry, StaticRetry
from dagorama.runner import execute, execute_async


class CustomNameDag(DAGDefinition):
    @dagorama(queue_name="test_queue").syncfn
    def entrypoint(self):
        return 10


class CustomAsyncDag(DAGDefinition):
    @dagorama(queue_name="test_queue").asyncfn
    async def entrypoint(self):
        return 10


class CustomTaintDag(DAGDefinition):
    @dagorama(taint_name="test_taint").syncfn
    def entrypoint(self):
        return 10


class CustomStaticErroringDag(DAGDefinition):
    @dagorama(retry=StaticRetry(max_attempts=2, interval=1)).syncfn
    def entrypoint(self):
        raise ValueError()


class CustomExponentialErroringDag(DAGDefinition):
    @dagorama(retry=ExponentialRetry(max_attempts=2, base_interval=2)).syncfn
    def entrypoint(self):
        raise ValueError()


class ThirdPartyClass:
    def __init__(self):
        self.constant = 1

    def __call__(self):
        return self.constant

    def other_fn(self):
        return self.constant


class CustomThirdPartyDag(DAGDefinition):
    def __init__(self):
        self.third_party = ThirdPartyClass()
        # self.other_fn = self.third_party.other_fn

    @dagorama().syncfn
    def entrypoint(self):
        return self.third_party()
        # Isn't currently supported
        # self.other_fn()


def test_custom_name(broker):
    dag = CustomNameDag()
    dag_instance, dag_result = dag()

    # Should run this queue
    execute(
        include_queues=["test_queue"],
        infinite_loop=False,
        catch_exceptions=False,
    )
    assert resolve(dag_instance, dag_result) == 10

    dag = CustomNameDag()
    dag_instance, dag_result = dag()

    # Should not run this queue
    execute(
        include_queues=["test_queue_2"],
        infinite_loop=False,
        catch_exceptions=False,
    )
    assert resolve(dag_instance, dag_result) == None


@pytest.mark.asyncio
async def test_async(broker):
    dag = CustomAsyncDag()
    dag_instance, dag_result = await dag()

    # Should run this queue
    await execute_async(
        include_queues=["test_queue"],
        infinite_loop=False,
        catch_exceptions=False,
    )
    assert resolve(dag_instance, dag_result) == 10


def test_taint_name(broker):
    dag = CustomTaintDag()
    dag_instance, dag_result = dag()

    # Should not run a tained queue by default
    execute(infinite_loop=False, catch_exceptions=False)
    assert resolve(dag_instance, dag_result) == None

    # Require specific allowance
    execute(queue_tolerations=["test_taint"], infinite_loop=False, catch_exceptions=False)
    assert resolve(dag_instance, dag_result) == 10


@pytest.mark.parametrize("dag_class", [CustomStaticErroringDag, CustomExponentialErroringDag])
def test_erroring_dags(broker, dag_class):
    dag = dag_class()
    dag_instance, dag_result = dag()

    # Should not run a tained queue by default
    execute(infinite_loop=False, catch_exceptions=True)
    assert resolve(dag_instance, dag_result) == None


def test_custom_third_party_dag(broker):
    """
    Ensure that instance variables are callable
    """
    dag = CustomThirdPartyDag()
    dag_instance, dag_result = dag()

    execute(infinite_loop=False, catch_exceptions=True)
    assert resolve(dag_instance, dag_result) == 1



from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional
from uuid import UUID

if TYPE_CHECKING:
    from dagorama.models.arguments import DAGArguments


@dataclass
class DAGPromise:
    """
    A promise of a future DAG result. These promises values are not directly
    usable locally but can be passed to other DAG functions. The runner will ensure
    that they are fully realized before the DAG function is called.

    """
    identifier: UUID

    function_name: str | None = None

    arguments: Optional["DAGArguments"] = None



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import abc
import dagorama.api.api_pb2
import grpc

class DagoramaStub:
    def __init__(self, channel: grpc.Channel) -> None: ...
    CreateWorker: grpc.UnaryUnaryMultiCallable[
        dagorama.api.api_pb2.WorkerConfigurationMessage,
        dagorama.api.api_pb2.WorkerMessage,
    ]
    CreateInstance: grpc.UnaryUnaryMultiCallable[
        dagorama.api.api_pb2.InstanceConfigurationMessage,
        dagorama.api.api_pb2.InstanceMessage,
    ]
    CreateNode: grpc.UnaryUnaryMultiCallable[
        dagorama.api.api_pb2.NodeConfigurationMessage,
        dagorama.api.api_pb2.NodeMessage,
    ]
    Ping: grpc.UnaryUnaryMultiCallable[
        dagorama.api.api_pb2.WorkerMessage,
        dagorama.api.api_pb2.PongMessage,
    ]
    GetNode: grpc.UnaryUnaryMultiCallable[
        dagorama.api.api_pb2.NodeRetrieveMessage,
        dagorama.api.api_pb2.NodeMessage,
    ]
    GetWork: grpc.UnaryUnaryMultiCallable[
        dagorama.api.api_pb2.WorkerMessage,
        dagorama.api.api_pb2.NodeMessage,
    ]
    SubmitWork: grpc.UnaryUnaryMultiCallable[
        dagorama.api.api_pb2.WorkCompleteMessage,
        dagorama.api.api_pb2.NodeMessage,
    ]
    SubmitFailure: grpc.UnaryUnaryMultiCallable[
        dagorama.api.api_pb2.WorkFailedMessage,
        dagorama.api.api_pb2.NodeMessage,
    ]

class DagoramaServicer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def CreateWorker(
        self,
        request: dagorama.api.api_pb2.WorkerConfigurationMessage,
        context: grpc.ServicerContext,
    ) -> dagorama.api.api_pb2.WorkerMessage: ...
    @abc.abstractmethod
    def CreateInstance(
        self,
        request: dagorama.api.api_pb2.InstanceConfigurationMessage,
        context: grpc.ServicerContext,
    ) -> dagorama.api.api_pb2.InstanceMessage: ...
    @abc.abstractmethod
    def CreateNode(
        self,
        request: dagorama.api.api_pb2.NodeConfigurationMessage,
        context: grpc.ServicerContext,
    ) -> dagorama.api.api_pb2.NodeMessage: ...
    @abc.abstractmethod
    def Ping(
        self,
        request: dagorama.api.api_pb2.WorkerMessage,
        context: grpc.ServicerContext,
    ) -> dagorama.api.api_pb2.PongMessage: ...
    @abc.abstractmethod
    def GetNode(
        self,
        request: dagorama.api.api_pb2.NodeRetrieveMessage,
        context: grpc.ServicerContext,
    ) -> dagorama.api.api_pb2.NodeMessage: ...
    @abc.abstractmethod
    def GetWork(
        self,
        request: dagorama.api.api_pb2.WorkerMessage,
        context: grpc.ServicerContext,
    ) -> dagorama.api.api_pb2.NodeMessage: ...
    @abc.abstractmethod
    def SubmitWork(
        self,
        request: dagorama.api.api_pb2.WorkCompleteMessage,
        context: grpc.ServicerContext,
    ) -> dagorama.api.api_pb2.NodeMessage: ...
    @abc.abstractmethod
    def SubmitFailure(
        self,
        request: dagorama.api.api_pb2.WorkFailedMessage,
        context: grpc.ServicerContext,
    ) -> dagorama.api.api_pb2.NodeMessage: ...

def add_DagoramaServicer_to_server(servicer: DagoramaServicer, server: grpc.Server) -> None: ...



from abc import ABC, abstractmethod
from contextlib import contextmanager
from functools import wraps
from inspect import isawaitable, ismethod
from pickle import loads
from typing import Any, Awaitable, cast
from uuid import UUID, uuid4

import grpc

import dagorama.api.api_pb2 as pb2
import dagorama.api.api_pb2_grpc as pb2_grpc
from dagorama.models.promise import DAGPromise

LAUNCH_RETURN = tuple["DAGInstance", DAGPromise]


class DAGDefinition(ABC):
    def __call__(self, *args, **kwargs) -> LAUNCH_RETURN | Awaitable[LAUNCH_RETURN]:
        """
        If entrypoint is an async function, will return an awaitable value. If it is a sync
        function will return immediately.

        """
        # We don't have an instance ID yet for this invocation
        instance_id = generate_instance_id()
        instance = DAGInstance(instance_id, self)

        # We want to run the first function (entrypoint) as part of the
        # broader instance context
        result_promise = instance.entrypoint(*args, **kwargs)

        if isawaitable(result_promise):
            return self.call_async(instance, result_promise)
        return self.call_sync(instance, result_promise)

    def call_sync(self, instance: "DAGInstance", result_promise: DAGPromise):
         # Cast as an actual result promise since this is what clients expect
        result_promise = cast(DAGPromise, result_promise)

        return instance, result_promise

    async def call_async(self, instance: "DAGInstance", promise: Awaitable[DAGPromise]):
        result_promise = await promise

        # Cast as an actual result promise since this is what clients expect
        result_promise = cast(DAGPromise, result_promise)

        return instance, result_promise

    @abstractmethod
    def entrypoint(self, *args, **kwargs):
        pass


class DAGInstance:
    def __init__(self, instance_id: UUID, definition: DAGDefinition):
        self.instance_id = instance_id
        self.definition = definition

    def __getattr__(self, name: str) -> Any:
        value = getattr(self.definition, name)
        if callable(value):
            value = inject_instance(self)(value)
            return value
        return getattr(self.definition, name)

    def __setattr__(self, name: str, value: Any) -> None:
        if name in ["instance_id", "definition"]:
            return super().__setattr__(name, value)
        return setattr(self.definition, name, value)

    def __delattr__(self, name: str) -> None:
        if name in ["instance_id", "definition"]:
            return super().__delattr__(name)
        return delattr(self.definition, name)


@contextmanager
def dagorama_context():
    # TODO: Get global context otherwise creates it

    with grpc.insecure_channel("localhost:50051") as channel:
        yield pb2_grpc.DagoramaStub(channel)


def generate_instance_id() -> UUID:
    # Calling indicates that we should spin off a new DAG instance
    with dagorama_context() as context:
        instance_id = uuid4()
        context.CreateInstance(
            pb2.InstanceConfigurationMessage(
                identifier=str(instance_id)
            )
        )
    return instance_id


def inject_instance(instance):
    def decorator(func):
        # Deal with instance methods that inject their own "self" into the
        # function as part of the call execution
        if ismethod(func):
            func = getattr(func, "__func__")

            @wraps(func)
            def wrapper(*args, **kwargs):
                return func(instance, *args, **kwargs)
        else:
            # We only want to inject the instance into functions of this main DAG
            return func
        return wrapper

    return decorator


def resolve(instance: DAGInstance, promise: DAGPromise):
    """
    Given a promise (typically of the initial_entrypoint), will recursively resolve
    promises in a return value chain.

    ie. Promise A -> Promise B -> Promise C will shortcut to Promise C, which will
    then return the true value.

    """
    with dagorama_context() as context:
        current_return_value = promise

        while isinstance(current_return_value, DAGPromise):
            node = context.GetNode(
                pb2.NodeRetrieveMessage(
                    instanceId=str(instance.instance_id),
                    identifier=str(current_return_value.identifier),
                )
            )

            if not len(node.resolvedValue):
                return None

            resolved = loads(node.resolvedValue)
            if resolved is None:
                return None

            current_return_value = resolved

        return current_return_value



from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from functools import wraps
from inspect import iscoroutinefunction
from logging import warning
from typing import Any, ParamSpec, TypeVar, cast
from uuid import uuid4

import dagorama.api.api_pb2 as pb2
from dagorama.code_signature import calculate_function_hash
from dagorama.definition import DAGDefinition, DAGInstance, dagorama_context
from dagorama.inspection import find_promises
from dagorama.models.arguments import DAGArguments
from dagorama.models.promise import DAGPromise
from dagorama.retry import RetryConfiguration
from dagorama.serializer import function_to_name

T = TypeVar('T')
P = ParamSpec('P')


@dataclass
class WrapperResults:
    result: Any | None = None
    promise: DAGPromise | None = None


class dagorama:
    """
    The dagorama decorator wraps a distinct piece of work that should be separately
    queued in a background function. It is polymorphic with regard to async or sync functions and there
    are a few different ways to execute it.

    The first and easiest is to use a vanilla `@dagorama` decorator. This will detect if the underlying
    wrapped function is async or sync and will queue it accordingly. This isn't strictly recommended
    because it results in a vaguer typing signature since we can't statically determine whether
    we are dealing with an async or sync function [1].

    ```
    @dagorama()
    def my_function():
        pass
    ```

    For more specific typing, call the `dagorama` decorator with either `dagorama.syncfn` or `dagorama.asyncfn`:

    ```
    @dagorama().syncfn
    def my_function():
        pass

    @dagorama().asyncfn
    async def my_function():
        pass
    ```

    The actual return type of functions wrapped with @dagorama() will be a DAGPromise. This is not what we want during
    development because this is a typeless type. Instead, we want to program the graph as if all promises are instantly
    fulfilled and the values are passed downstream.

    [1] The issue here is one of parity. We want the input type signature to match the output type signature explicitly
    to guarantee that we're correctly mocking the DAGPromise output. Sync functions result in a `T` output type whereas
    async functions result in an `Awaitable[T]` output. It's not currently possible to have the specification of one input
    type result in the determination of the output type.

    :param taint: When a taint is provided, workers will need to explicitly provide a toleration to execute. This is used
        in cases where there are specific compute resources that should own one phase of the pipeline.

    """
    def __init__(
        self,
        queue_name: str | None = None,
        taint_name: str | None = None,
        retry: RetryConfiguration | None = None,
    ):
        self.queue_name = queue_name
        self.taint_name = taint_name
        self.retry = retry

    def __call__(self, func: Callable[P, T | Awaitable[T]]):
        warning("Unknown type signature with `@dagorama` decorator. Use `@dagorama.sync` or `@dagorama.async` for explicit typing.")
        if iscoroutinefunction(func):
            return self.syncfn(func)
        else:
            return self.asyncfn(func)

    @property
    def syncfn(self):
        def decorator(func: Callable[P, T]) -> Callable[P, T]:
            if iscoroutinefunction(func):
                raise ValueError("dagorama().async is required for async functions")

            @wraps(func)
            def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                payload = self.common_wrapper(func, *args, **kwargs)
                if payload.result:
                    return payload.result

                # Wrong cast of types but we want the static typechecker to believe that the function
                # is returning the actual value as specified by the client caller
                # https://docs.python.org/3/library/typing.html#typing.ParamSpec
                return cast(T, payload.promise)

            wrapper.original_fn = func  # type: ignore
            return wrapper
        return decorator

    @property
    def asyncfn(self):
        def decorator(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
            if not iscoroutinefunction(func):
                raise ValueError("dagorama.sync() is required for sync functions")

            @wraps(func)
            async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                payload = self.common_wrapper(func, *args, **kwargs)
                if payload.result:
                    return await payload.result

                return cast(
                    T, payload.promise,
                )

            wrapper.original_fn = func  # type: ignore
            return wrapper
        return decorator

    def common_wrapper(
        self,
        func: Callable[P, T],
        *args: P.args,
        **kwargs: P.kwargs
    ):
        # Ignore the first argument, which will be the passed through class definition
        if isinstance(args[0], DAGDefinition):
            raise ValueError("A @dagorama() function should only be called as part of a DAGInstance")

        # We should instead be called on an instance of the DAG, which is injected into
        # the second argument slot
        if isinstance(args[0], DAGInstance):
            instance = args[0]
        else:
            raise ValueError("@dagorama can only be called on DAGInstances")

        # Can't be provided as an explicit keyword parameter because of a mypy constraint with P.kwargs having
        # to capture everything
        # https://github.com/python/typing/discussions/1191
        greedy_execution = kwargs.pop("greedy_execution", False)
        if greedy_execution:
            result = func(*args, **kwargs)
            return WrapperResults(result=result)

        # Strip out the class definition when we store the arguments
        isolated_args = cast(list, args)[1:]

        # This function will have a result
        # Queue in the DAG backend
        promise = DAGPromise(
            uuid4(),
            function_to_name(func),
            DAGArguments(
                isolated_args,
                kwargs
            )
        )

        # Find the dependencies
        promise_dependencies = find_promises([isolated_args, kwargs])

        # Add to the remote runloop
        with dagorama_context() as context:
            context.CreateNode(
                pb2.NodeConfigurationMessage(
                    identifier=str(promise.identifier),
                    functionName=cast(str, promise.function_name),
                    functionHash=calculate_function_hash(func),
                    taintName=self.taint_name or "",
                    queueName=self.queue_name or cast(str, promise.function_name),
                    arguments=(
                        cast(
                            # We know this is a valid argument object because we just set it
                            DAGArguments,
                            promise.arguments,
                        )
                        .to_server_bytes()
                    ),
                    sourceIds=[
                        str(dependency.identifier)
                        for dependency in promise_dependencies
                    ],
                    instanceId=str(instance.instance_id),
                    retry=self.retry.as_message() if self.retry is not None else None,
                )
            )

        return WrapperResults(promise=promise)
